package de.rcr1.brain;

import java.io.*;
import java.util.*;

/**
 * The intelligence to guide the robot.
 * 
 * @author Kilian Koller and Philipp Hergert
 */
class Brain
{
    /** The Robot to move. */
    private final Robot pathFinder;
    
    /** The virtual map of the mars. */
    private final Board marsMap;
    
    /** The position of the robot. */
    private Node position;
    
    /** The open and closed list for the A*.  */
    private final Vector<Node> open;
    private final Vector<Node> closed;
    
    /**
     * Standard ctor.
     * @throws IOException
     * @throws FileNotFoundException
     * @throws InvalidPropertiesFormatException
     */
    Brain() throws InvalidPropertiesFormatException, FileNotFoundException, IOException
    {
        // Read properties file
        Properties props = new Properties();
        props.loadFromXML(new FileInputStream("host.xml"));
        
        System.out.println("Connecting to: " + props.getProperty("ip") + ":" + Integer.parseInt(props.getProperty("port")));
        
        pathFinder = new RemoteRobot(props.getProperty("ip"), Integer.parseInt(props.getProperty("port")));
        
        marsMap = new Board();
        open = new Vector<Node>();
        closed = new Vector<Node>();
        
        // Start in the top left corner.
        position = new Node(0, 0);
    }
    
    /**
     * Uses the A* algorith to generate a vector with robot commands.
     * 
     * @return a command list.
     */    
    Vector<Integer> calculatePath()
    {
        // Start position.
        Node start = new Node(position);
        
        // Goal position.
        Node goal = new Node(Constants.MARS_WIDTH - 1, Constants.MARS_HEIGHT - 1);
        
        // The node to walk back from when done.
        Node lastSuccessor = null;
        
        // Add start to open list.
        open.add(start);
        
        while(open.size() > 0)
        {
            // Set the node with the lowest f from open to be the current node.
            Node current = findLowestFNode(open, goal);
            
            // Remove current from open.
            open.remove(current);
            
            // Get the possible surrounding nodes of current.
            Vector<Node> surroundingPositions = marsMap.getSurroundingPositions(current);
            
            // And add them to open if:
            for(int i = 0; i < surroundingPositions.size(); i++)
            {
            	Node successor = surroundingPositions.get(i);
            	
                // If current equals goal we are done.
                if(successor.equalPosition(goal))
                {
                    // Empty lists for next calculation
                    open.removeAllElements();
                    closed.removeAllElements();
                    
                    // Set lastSuccessor to be the last node found. 
                    lastSuccessor = successor;
                    break;
                }
                
                // If open contains this position and has a faster way(lower F), drop this way.
                Node temp = successor.getContained(open); // successor from openlist with alternate route 
                if(temp != null &&
                		temp.getF(goal) <= successor.getF(goal))
                    continue;

                // If closed contains this position and has a faster way(lower F), drop this way.
                temp = successor.getContained(closed); // successor from closedlist with alternate route 
                if(temp != null &&
                		temp.getF(goal) <= successor.getF(goal))
                    continue;
                
                // Add successor to open.
                open.add(successor);
            }
            
            // Add current to closed.
            closed.add(current);
        }

        // Reverse the path.
        Node next = lastSuccessor;
        Vector<Node> reversePath = new Vector<Node>();
        while(next != null)
        {
            reversePath.add(0, next);
            next = next.getParent();
        }
        
        if(Constants.DEBUG)
        {
        // Print the reversed path.
        	System.out.println("-------------------------");
        	System.out.println("Path backwards:");
        	for(Node n: reversePath)
        	{
        		System.out.println(n.toString(goal));
        	}
        }
        
        // Generate command list for robot.
        Vector<Integer> commands = new Vector<Integer>();
        for(int i = 0; i < reversePath.size() - 1; i++)
        {
            // Horizontal - right.
            if(reversePath.get(i + 1).getPosH() - reversePath.get(i).getPosH() == 1)
                commands.add(new Integer(Constants.OPCODE_RIGHT));
            // Horizontal - left.
            else if(reversePath.get(i + 1).getPosH() - reversePath.get(i).getPosH() == -1)
                commands.add(new Integer(Constants.OPCODE_LEFT));
            // Vertical - down.
            else if(reversePath.get(i + 1).getPosV() - reversePath.get(i).getPosV() == 1) 
            {
                //System.err.println("FORWARD: " + Constants.OPCODE_FORWARD);
            	commands.add(new Integer(Constants.OPCODE_FORWARD));
            	
            // Vertical - up.
            } 
            else if(reversePath.get(i + 1).getPosV() - reversePath.get(i).getPosV() == -1)
                commands.add(new Integer(Constants.OPCODE_BACKWARD));
        }
        
        if(Constants.DEBUG)
        {
        	// Print generated commands.
        	System.out.println("-------------------------");
        	System.out.println("Commands (see Constants):");
        	for(Integer c: commands)
        	{
        		printCommand(c.intValue());
        	}
        	System.out.println();
        }
        
        // Return the commands.
        return commands;
    }
    
    /**
     * Prints a command on stdOut.
     * @param c command to print
     */
    void printCommand(int c)
    {
    	System.out.print("command: ");
        switch(c)
		{
	        case Constants.OPCODE_FORWARD:
	        	System.out.println("forward");
	        	break;
	        case Constants.OPCODE_BACKWARD:
	        	System.out.println("backward");
	        	break;
	        case Constants.OPCODE_RIGHT:
	        	System.out.println("right");
	        	break;
	        case Constants.OPCODE_LEFT:
	        	System.out.println("left");
	        	break;
	        default:
	        	System.out.println("bad command.");
	        	break;
		}
    }
    
    /**
     * Groups the first commands in the list with the same direction.
     * 
     * @param commands the commands to group.
     * @return ammount of commands in the same direction. 
     */
    int groupCommands(Vector<Integer> commands)
    {
    	if (commands.size() <= 1)
    		return commands.size();

    	int ammount = 1;

    	for(int i = 0; i < commands.size() - 1; i++)
    	{
    		if(commands.get(i).intValue() == commands.get(i+1).intValue())
    			ammount++;
    		else
    			break;
    	}
    	
    	return ammount;
    }
    
    /**
     * Rides along the mars.
     */
    void ride()
    {
        // Calculate the path;
        Vector<Integer> path = calculatePath();

		while(path.size() > 0)
		{
	        int command = path.firstElement().intValue();
		    int distance = groupCommands(path) * Constants.MARS_GRID_SIZE;

		    System.out.println("Sending: " + command + "/" + distance);
		    Packet receive = pathFinder.sendCommand(new Packet(command, distance));
		    int realDistance = receive.getValue();
		    
		    System.out.println("Moved " + realDistance + "/" + distance);
	        
	        // ride back into grid
	        int backCommand = 0;
	        switch(command)
	        {
		    	case Constants.OPCODE_FORWARD:
		    		backCommand = Constants.OPCODE_BACKWARD;
		        	break;
		        case Constants.OPCODE_BACKWARD:
		    		backCommand = Constants.OPCODE_FORWARD;
		        	break;
		        case Constants.OPCODE_RIGHT:
		        	backCommand = Constants.OPCODE_LEFT;
		        	break;
		        case Constants.OPCODE_LEFT:
		        	backCommand = Constants.OPCODE_RIGHT;
		        	break;		        
	        }
	        pathFinder.sendCommand(new Packet(backCommand, realDistance % Constants.MARS_GRID_SIZE));
	        
	        // Remove all done commands
		    for (int i = 0; i < realDistance / Constants.MARS_GRID_SIZE; i++)
		    {
		        // Set the new position of the robot.
		        switch(command)
		        {
			    	case Constants.OPCODE_FORWARD:
			        	position = new Node(position.getPosH(), position.getPosV() + 1);
			        	break;
			        case Constants.OPCODE_BACKWARD:
			            position = new Node(position.getPosH(), position.getPosV() - 1);
			        	break;
			        case Constants.OPCODE_RIGHT:
			            position = new Node(position.getPosH() + 1, position.getPosV());
			        	break;
			        case Constants.OPCODE_LEFT:
			            position = new Node(position.getPosH() - 1, position.getPosV());
			        	break;		        
		        }
		        path.remove(0);
		    }
		    
		    System.out.println("New robot position: " + position.getPosH() + "/" + position.getPosV());
		    
		    // Couldn't walk the requested distance.
		    if(realDistance != distance)
		    {
		        // Mark blocked way.
		        int pathFinderH = position.getPosH();
		        int pathFinderV = position.getPosV();
		        
		        switch(command)
		        {
			    	case Constants.OPCODE_FORWARD:
			    	    marsMap.setBlocked(new Node(pathFinderH, pathFinderV + 1));
			        	break;
			        case Constants.OPCODE_BACKWARD:
			            marsMap.setBlocked(new Node(pathFinderH, pathFinderV - 1));
			        	break;
			        case Constants.OPCODE_RIGHT:
			            marsMap.setBlocked(new Node(pathFinderH + 1, pathFinderV));
			        	break;
			        case Constants.OPCODE_LEFT:
			            marsMap.setBlocked(new Node(pathFinderH - 1, pathFinderV));
			        	break;		        
		        }

		        if(Constants.DEBUG)
		        {
		        	// Print map.
		        	System.out.println("Obstacle encountered.");
		        	System.out.println("-------------------------");
		        	System.out.println("Virtual map:");
		        	marsMap.print();
		        }
		        	
		        // Recalculate the path and restart the loop.
		        path = calculatePath();
		    }
		}
		
		// Afterburner
		pathFinder.sendCommand(new Packet(Constants.OPCODE_FORWARD, 30));
		pathFinder.sendCommand(new Packet(Constants.OPCODE_RIGHT, 30));
    }
    
    /**
     * Finds the node with the lowest f from a list of nodes.
     * 
     * @param nodes the node list to find the node in.
     * @param goal the goal to calculate f with.
     * @return the node with lowest f value.
     */
    Node findLowestFNode(Vector<Node> nodes, Node goal)
    {
        int minF = Integer.MAX_VALUE;
        Node lowestFNode = null;
        
        for(int i = 0; i < nodes.size(); i++)
        {
            // g for node i.
            int g = nodes.get(i).getG();
            
            // h for node i.
            int h = nodes.get(i).getH(goal);
            
            // Calculate f for node i.
            int f = g + h;
            
            // if f for node i lower than for lowestFNode node i is new lowestFNode.
            if(f < minF)
            {
                minF = f;
                lowestFNode = nodes.get(i);
            }
        }
        
        return lowestFNode;
    }

    /**
     * Main function, runs run.
     * 
     * @param args not needed.
     * @throws IOException
     * @throws FileNotFoundException
     * @throws InvalidPropertiesFormatException
     */
    public static void main(String[] args) throws InvalidPropertiesFormatException, FileNotFoundException, IOException
    {
        new Brain().ride();
    }
}
