package de.rcr1.brain;

/**
 * Represents the robot.
 * 
 * @author Kilian Koller and Philipp Hergert
 */
class LocalRobot implements Robot
{
    /** The real mars. */
    private final Board marsReal;
    
    /** The position of the robot. */
    private Node position;
    
    /**
     * Standard Ctor.
     */
    LocalRobot()
    {
        marsReal = new Board(
        new byte[][]
            {
                {0, 1, 0, 0, 0, 1, 0, 0, 0, 0},
                {0, 1, 0, 1, 0, 1, 0, 1, 0, 0},
                {0, 1, 0, 1, 0, 0, 0, 1, 0, 1},
                {0, 0, 0, 1, 0, 1, 0, 0, 1, 1},
                {0, 1, 0, 1, 0, 1, 1, 0, 0, 0}
            }
        );
   
        
        // Start in the top left corner.
        position = new Node(0, 0);
    }
    
    /**
     * @see de.rcr1.brain.Robot#sendCommand(de.rcr1.brain.Packet)
     */
    public Packet sendCommand(Packet p)
    {
        int distance = p.getValue();
        int realDistance = move(p.getCommand(),distance);
        int returnCommand;
        
        if (distance != realDistance)
        	returnCommand = Constants.OPCODE_MOVE_ERROR;
        else
        	returnCommand = Constants.OPCODE_MOVE_SUCCESS;
        
    	return new Packet(returnCommand, realDistance);
    }
    
    /**
     * Moves the robot distance d in the negative vertical direction.
     * Move Down.
     * 
     * @param d the distance to move.
     * @return how far it went.
     */
    private int moveForward(int d)
    {
        int realDistance = 0;
        
        while((position.getPosV() < Constants.MARS_HEIGHT - 1) &&
                realDistance < d &&
                (marsReal.getPosition(new Node(position.getPosH(), position.getPosV() + 1)) != Constants.FIELD_BLOCKED))
        {
            position = new Node(position.getPosH(), position.getPosV() + 1);
            realDistance++;
        }
        
        return realDistance;
    }
    
    /**
     * Moves the robot distance d in the postitive vertical direction.
     * Move Up.
     * 
     * @param d the distance to move.
     * @return how far it went.
     */
    private int moveBackward(int d)
    {
        int realDistance = 0;
        
        while((position.getPosV() > 0) &&
                realDistance < d &&
                (marsReal.getPosition(new Node(position.getPosH(), position.getPosV() - 1)) != Constants.FIELD_BLOCKED))
        {
            position = new Node(position.getPosH(), position.getPosV() - 1);
            realDistance++;
        }
        
        return realDistance;
    }
    
    /**
     * Moves the robot distance d in the postitive horizontal direction.
     * Move Right.
     * 
     * @param d the distance to move.
     * @return how far it went.
     */
    private int moveRight(int d)
    {
        int realDistance = 0;
        
        while((position.getPosH() < Constants.MARS_WIDTH - 1) &&
                realDistance < d &&
                (marsReal.getPosition(new Node(position.getPosH() + 1, position.getPosV())) != Constants.FIELD_BLOCKED))
        {
            position = new Node(position.getPosH() + 1, position.getPosV());
            realDistance++;
        }
        
        return realDistance;
    }
    
    /**
     * Moves the robot distance d in the negative horizontal direction.
     * Move Left.
     * 
     * @param d the distance to move.
     * @return how far it went.
     */
    private int moveLeft(int d)
    {
        int realDistance = 0;
    
	    while((position.getPosH() > 0) &&
	            realDistance < d &&
	            (marsReal.getPosition(new Node(position.getPosH() - 1, position.getPosV())) != Constants.FIELD_BLOCKED))
	    {
	        position = new Node(position.getPosH() - 1, position.getPosV());
	        realDistance++;
	    }
	    
	    return realDistance;     
    }
    
    /**
     * Move using opcodes.
     * 
     * @param c the command opcode.
     * @param d the distance to move.
     * 
     * @return how far it went.
     */
    private int move(int c, int d)
    {
        switch (c)
        {
	    	case Constants.OPCODE_FORWARD:
	    	    return moveForward(d);
	        case Constants.OPCODE_BACKWARD:
	            return moveBackward(d);
	        case Constants.OPCODE_RIGHT:
	            return moveRight(d);
	        case Constants.OPCODE_LEFT:
	            return moveLeft(d);
	        default:
	            return 0;
        }   
    }
}
