package de.rcr1.brain;

import java.util.Vector;

/**
 * This class is used for A* ans as a position helper class.
 * 
 * @author Kilian Koller
 */
public class Node
{
    /** The position in the horizontal axis. */
    private int posH;
  
    /** The position in the vertical axis. */
    private int posV;
    
    /** The parent node of this node. */
    private Node parent;
    
    /** The distance from the start. */
    private int gdist;
    
    /**
     * Standard ctor.
     * 
     * @param h the horizontal position.
     * @param v the vertical position.
     * @param p the parent of the new node.
     * @param g the g distance of the new node.
     */
    Node(int h, int v, Node p, int g)
    {
        posH = h;
        posV = v;
        parent = p;
        gdist = g; 
    }
    
    /**
     * Ctor, without parent and g distance.
     * 
     * @param h the horizontal position.
     * @param v the vertical position.
     */
    Node(int h, int v)
    {
        this(h, v, null, 0);
    }
    
    /**
     * Copy ctor.
     * 
     * @param n the node to copy.
     */
    Node(Node n)
    {
        this(n.getPosH(), n.getPosV(), n.getParent(), n.getG());
    }
    
    /**
     * Calculates the Manhattan distance of this node to another.
     *
     * @param goal the node to calculate the distance to.
     * @return the Manhattan distance of this node to the other one.
     */  
    int getH(Node goal)
    {
        // Horizontal distance
        int distH = Math.abs(this.getPosH() - goal.getPosH());
        
        // Vertical distance
        int distV = Math.abs(this.getPosV() - goal.getPosV());
        
        return distH + distV;
    }
    
    /**
     * Calculates the distance from the start.
     * 
     * @return the distance from the start.
     */
    int getG()
    {
        return gdist;
    }
    
    /**
     * Returns the f value of this node.
     * 
     * @param goal the node to calculate the distance to.
     * @return the f value of this node.
     */
    int getF(Node goal)
    {
        return getG() + getH(goal);
    }
    
    /**
     * Returns the parent of this node.
     * 
     * @return the parent.
     */
    Node getParent()
    {
        return parent;
    }
    
    /**
     * Sets the parent of this node.
     * 
     * @param p the parent to set.
     */
    void setParent(Node p)
    {
        this.parent = p;
    }
    
    /**
     * @return Returns the posH.
     */
    int getPosH()
    {
        return posH;
    }
    
    /**
     * @return Returns the posV.
     */
    int getPosV()
    {
        return posV;
    }
    
    /**
     * Compares the position of this node with another one n.
     * 
     * @param n the node to compare the position to.
     * @return if the positions are equal.
     */
    boolean equalPosition(Node n)
    {
        return(posH == n.getPosH() && posV == n.getPosV());
    }
    
    /**
     * Returns a node from a node vector id the one give as a parameter has the same position.
     * 
     * @param nodelist a vactor of nodes.
     * @return a node fom the vector with position as this one.
     */
    Node getContained(Vector<Node> nodelist)
    {
        for(int i = 0; i < nodelist.size(); i++)
        {
            if(nodelist.get(i).posH == this.posH &&
                    nodelist.get(i).posV == this.posV)
                return nodelist.get(i);
        }
        
        return null;
    }
    
    /**
     * Builds a string representation of this node with g ang h values.
     * 
     * @param goal the goal node of the game.
     * @return a position string.
     */
    String toString(Node goal)
    {
    	return "(" + posH + ";" + posV + ") G: " + getG() + " H: " + getH(goal);
    }
}
