/**
 * Munich University of Applied Sciences, FB07 Computer Science
 * FWP - Remotely Controlled Robot - 7.Semester
 * 
 * Class Description:
 * 		This is the main class containing the main() method running
 * 		on the RCX.
 * 
  * Error Codes: (code - method - cause)
 * 		EA1 - run() - Received unknown opcode
 * 		EA2 - run() - Lost Connection to Proxy - Too many timeouts
 * 		EA3 - main(String[]) - Fatal Error, no way to handle, shutting down the RCX
 * 
 * @author Thomas Woellert (thomas.woellert@t-online.de)
 * @author Stefan Weissbach (weissbst@gmx.de)
 * @author Willi Tremmel (TremmelW@gmx.de)
 * 
 * System: Linux 2.6, Sun Java JDK 5.0
 * Semestergroup: 7T
 */
package rcx;

import josx.platform.rcx.Motor;
import josx.platform.rcx.Sensor;
import josx.platform.rcx.SensorConstants;
import josx.platform.rcx.Sound;
import josx.platform.rcx.TextLCD;
import josx.rcxcomm.LLC;
import util.IRMessage;
import util.IRProtocol;
import util.LostConnectionException;

public class RCXMain {
	
	/* Time till a timeout occurs on the proxy side */
	protected int timeout = 6000;
	
	/* How often should we retry after a timeout */
	protected int numberOfRetries = 3;
	
	/* The name of our robot */
	protected String name = "BENDER";
	
	/* The handler taking care of communication */
	protected RCXMessageHandler messageHandler = new RCXMessageHandler(timeout, numberOfRetries);
	
	/* The handler taking care of movement */
	protected RCXMovementHandler movementHandler = new RCXMovementHandler();
	
	/* Our own message object to send to and receive message from the tower */
	protected IRMessage message = null;

	/**
	 * Create a new RCXMain object
	 */
	public RCXMain() {
    	// Display the name of our robot in the LCD
    	TextLCD.print(name);
		
    	// Initiate the robot
    	this.init();
		
		// Get under way ...
		this.run();
		
		// Set to long range
		LLC.setRangeLong();
	}
	
	/**
	 * Initialize the robot
	 */
	protected void init() {
    	// Init the sensors
        Sensor.S1.setTypeAndMode(SensorConstants.SENSOR_TYPE_ROT, SensorConstants.SENSOR_MODE_ANGLE); // F/B
        Sensor.S2.setTypeAndMode(SensorConstants.SENSOR_TYPE_TOUCH, SensorConstants.SENSOR_MODE_BOOL); // TOUCH
        Sensor.S3.setTypeAndMode(SensorConstants.SENSOR_TYPE_ROT, SensorConstants.SENSOR_MODE_ANGLE); // R/L

        // Activate the rotational sensors
        Sensor.S1.activate();
        Sensor.S3.activate();
        
        // Init the engines
        Motor.A.setPower(7); // F/B
        Motor.B.setPower(7); // Hoch/Runter
        Motor.C.setPower(4); // R/L
	}
	
	/**
	 * This method performs the RCX task to wait for incoming
	 * messages and act according to the opcodes in these messages
	 */
	private void run() {
		// Give ready sound
		Sound.beep();
		
		// Run forever
		while(true) {
			
			try {
				// First we need to receive a message so we know what to do (blocks till a message has been received)
				this.message = this.messageHandler.receiveMessageFromTower();
				
				// Check if we received a message (the message is "null" if an error occured in the RCXMessageHandler)
				if(this.message != null) {
					
					// Take a look at the received opcode
					if(this.message.getOpcode() == IRProtocol.OPCODE_FORWARD) {
						
						// Drive forward and then send the given reply to the server
						this.message = this.movementHandler.forward(this.message.getValue());
						this.messageHandler.sendMessageToTower(this.message);
						
					} else if(this.message.getOpcode() == IRProtocol.OPCODE_BACKWARD) {
						
						// Drive backward and then send the given reply to the server
						this.message = this.movementHandler.backward(this.message.getValue());
						this.messageHandler.sendMessageToTower(this.message);
						
					} else if(this.message.getOpcode() == IRProtocol.OPCODE_RIGHT) {
						
						// Drive right and then send the given reply to the server
						this.message = this.movementHandler.right(this.message.getValue());
						this.messageHandler.sendMessageToTower(this.message);
						
					} else if(this.message.getOpcode() == IRProtocol.OPCODE_LEFT) {
						
						// Drive left and then send the given reply to the server
						this.message = this.movementHandler.left(this.message.getValue());
						this.messageHandler.sendMessageToTower(this.message);
						
					} else if(this.message.getOpcode() == IRProtocol.OPCODE_ACTIVE) {
						
						// Send the proper reply to the server
						this.messageHandler.sendMessageToTower(IRProtocol.OPCODE_ACTIVE);
					} else {
						
						// Display Error - Unknown Opcode
						Sound.buzz();
						TextLCD.print("EA1");
					}
				}
			} catch (LostConnectionException e) {
				// Display Error - Lost Connection to Proxy - Too many timeouts
				Sound.buzz();
				TextLCD.print("EA2");
			}
		}
	}
	
	/**
	 * Called when we start the RCX
	 * @param args, no args given
	 */
	public static void main(String[] args) {
		try {
			// Create the main object
			RCXMain main = new RCXMain();
		} catch (Exception e) {
			// Display Error - Fatal Error, shutting down the RCX
			Sound.buzz();
			TextLCD.print("EA3");
			Sound.buzz();
			
			System.exit(0);
		}
	}
}
