/**
 *                      Remotely Controlled Robot
 *                    FHM Munich FB07 Computer Science
 *                         WS2004/05 Group 1
 *                          Proxy Module
 *
 *
 * Class Description:
 *          This class is used for handling the communication with
 *          the ir-tower and the robot over a ir-connection
 *
 * System:
 *          Linux 2.6, Sun Java 1.4.2; Win32, Sun Java 1.4.2
 *
 * @author Stephan Poser <stephan.poser@gmx.de>
 *
 *
 * $Id: RcxMessageHandler.java 165 2004-12-09 18:51:31Z poser $
 *
 */

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.lang.reflect.Method;

import josx.rcxcomm.RCXPort;

/**
 * class used for communicating with the
 * ir-tower and the robot over a ir-connection
 *
 * @author tom, stephan
 *
 */
public class RcxMessageHandler extends AbstractMessageHandler {

	/* Time till a timeout occurs on the proxy side */
	protected static int timeout = 3000;

	/* How often should we retry after a timeout */
	protected static int numberOfRetries = 5;

	/* Show debug messages ? */
	protected static boolean isDebug = false;

	/* communication handler */
	ProxyMessageHandler messageHandler;

	/**
	 * Create the default message handler
	 */
	public RcxMessageHandler() {
		// Create the handler taking care of communication
		messageHandler = new ProxyMessageHandler(timeout, numberOfRetries, isDebug);

	}

	/**
	 * Send a specific Message to the RCX
	 * @param opcode, the IRProtocol opcode to send
	 * @param distance, the distance to send
	 */
	public void sendMessageToRCX(Packet packet) throws IllegalArgumentException, LostConnectionException {

		if(packet.getOpcode() == IRProtocol.OPCODE_ACTIVE)
				messageHandler.sendMessageToRCX(packet.getOpcode());
			else
				messageHandler.sendMessageToRCX(packet.getOpcode(), packet.getData());

		//wait for reply
		IRMessage message = messageHandler.receiveMessageFromRCX();
		fireMessageReceived(new Packet(message.getOpcode(), message.getValue()));
	}

    /**
     * Let all listeners know when a message has been received
     * @param opcode, the received opcode
     * @param distance, the received distance
     */
    protected void fireMessageReceived(Packet packet) {
    	// Go through the whole listener list
    	for(int i = 0; i < listenerList.size(); i++) {
    		((MessageListener)(listenerList.elementAt(i))).readFromDest(packet);
    	}
    }

//}

    /**
     * Munich University of Applied Sciences, FB07 Computer Science
     * FWP - Remotely Controlled Robot - 7.Semester
     *
     * Class Description:
     * 		This class takes care of receiving messages from the RCX and should be used
     * 		on the proxy side of the program. It offers two methods to send messages to the
     * 		RCX and one method to receive messages from the RCX when necessary.
     *
     * @author Thomas Woellert (thomas.woellert@t-online.de)
     * @author Stefan Weissbach (weissbst@gmx.de)
     * @author Willi Tremmel (TremmelW@gmx.de)
     *
     * System: Linux 2.6, Sun Java JDK 5.0
     * Semestergroup: 7T
     */


    class ProxyMessageHandler {

    	/* The port we use to communicate */
    	protected RCXPort port = null;

    	/* The stream for input communication */
    	protected DataInputStream dInputStream = null;

    	/* The stream for output communication */
    	protected DataOutputStream dOutputStream = null;

    	/* The object taking care of receiving messages from the tower */
    	protected ReceiveMessageHandler receiveMessageHandler = null;

    	/* Boolean telling us if we should show debug messages while running or not */
    	protected boolean isDebug = false;

    	/* The time to wait till timeout (in ms) */
    	protected int timeout = -1;

    	/* The number of retries to send a packet before final error */
    	protected int numberOfRetries = -1;
    	protected int retryCounter = 0;
    	protected int retryCounterACK = 0;

    	/* Variable to store received bytes in */
    	protected int receivedInt = -1;
    	protected int receivedIntACK = -1;

    	/* Boolean used to determine if all went well while sending bytes or if we got a RESEND request */
    	protected boolean transmissionOK = false;

    	/* Variables used when doing a timeout */
    	protected boolean isTimeout = false; // Boolean which is used by the subthreads to indicate a timeout
    	protected Thread timeoutThread = null; // The Thread object we use when writing something out onto the output stream
    	protected long startTime = -1; // The time when we started trying to send the message

    	/**
    	 * Create the proxy message handler
    	 * @param timeout, the value till a timeout occurs while reading or writing (in Milliseconds)
    	 * @param numberOfRetries, the number of retries after a timeout occured till the program is aborted
    	 * @param isDebug, true if debug messages should be writting to the console, false otherwise
    	 */
    	public ProxyMessageHandler(int timeout, int numberOfRetries, boolean isDebug) {

    		try {
    			// Store the values we received
    			this.timeout = timeout;
    			this.numberOfRetries = numberOfRetries;
    			this.isDebug = isDebug;

    			// Create the RCXPort object
    			port = new RCXPort();

    			// Get the input and output streams for this port
    			this.dInputStream = new DataInputStream(port.getInputStream());
    			this.dOutputStream = new DataOutputStream(port.getOutputStream());


    			// Create the handler to receive messages
    			this.receiveMessageHandler = new ReceiveMessageHandler();

    			if(isDebug) {
    				System.out.println("[Debug] - " + this.getClass().getName() + " - Running constructor complete");
    			}
    		} catch (Exception e) {
    			System.err.println("[Error] - " + this.getClass().getName() + " - constructor() - Error while creating object: " + e);
    		}
    	}

    	/**
    	 * Send a message from the IR tower to the RCX
    	 * @param opcode, the opcode to send
    	 * @throws IllegalArgumentException, thrown in case the input values are illegal
    	 * @throws LostConnectionException, thrown in case the number of timeouts reached the maximum number of retries
    	 */
    	public void sendMessageToRCX(int opcode) throws IllegalArgumentException, LostConnectionException {

    		// Check if the opcode the user wants to send is beyond our abilities: value must be <= 255
    		if(opcode > 8000 || opcode < 0) {

    			System.err.println("[Error] - " + this.getClass().getName() + " - sendMessageToRCX() - Can only send opcodes in the range 0-255. You tried (opcode): " + opcode);

    			throw new IllegalArgumentException();
    		} else {

    			if(isDebug) {
    				System.out.println("[Debug] - " + this.getClass().getName() + " - Sending message to RCX(opcode): " + opcode);
    			}

    			// Calling the proper method
    			this.sendMessageToRCX(opcode, 0);
    		}
    	}

    	/**
    	 * Send a packet with the given opcode and value to the RCX
    	 * @param opcode, the opcode to send
    	 * @param value, the value to send with the message (may be 0-255)
    	 * @throws IllegalArgumentException, thrown in case the input values are illegal
    	 * @throws LostConnectionException, thrown in case the number of timeouts reached the maximum number of retries
    	 */
    	public void sendMessageToRCX(int opcode, int value) throws IllegalArgumentException, LostConnectionException {

    		// Check if the value and opcode the user wants to send is beyond our abilities: value and opcode must be 0-255
    		// Value is allowed to be -1 in case we have no value to send, then only the opcode is transmitted
    		if(value > 999 || value < -1 || opcode > 8000 || opcode < 0) {

    			System.err.println("[Error] - " + this.getClass().getName() + " - sendMessageToRCX() - Can only send values or opcodes in the range 0-255. You tried (value/opcode): " + opcode + " / " + value);

    			throw new IllegalArgumentException();
    		} else {

    			if(isDebug) {
    				System.out.println("[Debug] - " + this.getClass().getName() + " - Sending message to RCX(opcode, value): " + opcode + " / " + value);
    			}

    			// Boolean is set to false for the while to run at least once
    			this.transmissionOK = false;

    			// Do again in case we have a transmission error
    			while(!this.transmissionOK) {

    				// Send the opcode + value
    				this.transmissionOK = this.sendValueToRCX((opcode + value));

    			}
    		}
    	}

    	/**
    	 * Causes the handler to wait for a message from the RCX
    	 * Blocks till a message has been received
    	 * @return the received message
    	 * @throws LostConnectionException, thrown in case the number of timeouts reached the maximum number of retries
    	 */
    	public IRMessage receiveMessageFromRCX() throws LostConnectionException {
    		if(isDebug) {
    			System.out.println("[Debug] - " + this.getClass().getName() + " - receiveMessageToRCX(): Waiting for a message from the RCX");
    		}
    		return this.receiveMessageHandler.read();
    	}

    	/**
    	 * Should be called when we want to write a value to the outputstream.
    	 * Tries to send the message to the RCX via the outputstream by creating
    	 * a new subthread. This method also kills the subthread if a timeout happens
    	 * and then throws an IOException so the calling method get's notified about the
    	 * timeout
    	 * @param value, the value to send to the RCX (0-255);
    	 * @throws IOException
    	 */
    	private void writeToOutputStream(int value) throws IOException {

    		if(isDebug) {
    			System.out.println("[Debug] - " + this.getClass().getName() + " - writeToOutputStream(): Writing to output Stream, value: " + value);
    		}

    		// Create a new Thread and start it
    		this.timeoutThread = new Thread(new SendMessageHandler(value));
    		this.timeoutThread.start();

    		// Remember the current system time as start time and set the boolean value so we get notified of the timeout
    		this.startTime = System.currentTimeMillis();
    		this.isTimeout = true;

    		if(isDebug) {
    			System.out.println("[Debug] - " + this.getClass().getName() + " - writeToOutputStream(): Waiting for Timeout: " + timeout + " ms");
    		}

    		// Stay in the while till the timeout time is reached or a timeout occured
    		while(System.currentTimeMillis() < (startTime + timeout) && this.isTimeout) {
    			// Do nothing
    		}

    		// Take a look if we stopped the while due to a timeout occuring
    		if(this.isTimeout) {

    			// Timeout took place
    			System.err.println("[Error] - " + this.getClass().getName() + " - writeToOutputStream(): Timeout occured. RCX not in range or inactive?");

    			// Bomb the subthread to the ground (ONLY WAY OF STOPPING IT!!!)
    			try {
    				if(isDebug) {
    					System.out.println("[Debug] - " + this.getClass().getName() + " - writeToOutputStream(): Bombing subthread to the ground");
    				}

    				// Grab the native "stop0()" method implemented in the Thread class
    				Method method = Thread.class.getDeclaredMethod( "stop0", new Class[] { Object.class } );
    				method.setAccessible( true );

    				// Initiate the method to kill the thread
    				method.invoke( this.timeoutThread, new Object[] { new ThreadDeath() } );

    			} catch (Exception e) {
    				System.err.println("[Error] - " + this.getClass().getName() + " - writeToOutputStream(): Error happened while bombing the subthread !");
    			}

    			// Timeout took place, throw a new IOException to notify the calling method of the error
    			throw new IOException();
    		}

    		// Set the thread object to null so that it may be collected by the garbage collector
    		this.timeoutThread = null;
    	}

    	/**
    	 * Used to send a value directly to the RCX
    	 * @param value, can be an opcode or a normal value
    	 * @return boolean, true if all went well, false if we got a resend request
    	 * @throws LostConnectionException, gets thrown in case that the number of timeouts reaches the maximum number of retries
    	 */
    	private boolean sendValueToRCX(int value) throws LostConnectionException {

    		if(isDebug) {
    			System.out.println("[Debug] - " + this.getClass().getName() + " - sendValueToRCX(): Sending value to RCX: " + value);
    		}

    		// Send the opcode till the RCX responds with an ACK
    		this.receivedInt = -1;
    		this.retryCounter = 0;

    		while(receivedInt != IRProtocol.OPCODE_MESSAGE_ACK && retryCounter <= this.numberOfRetries) {

    			try {
    				if(isDebug) {
    					System.out.println("[Debug] - " + this.getClass().getName() + " - sendValueToRCX(): Starting to send value");
    				}

    				// Send the value to the RCX with a timeout
    				this.writeToOutputStream(value);

    				if(isDebug) {
    					System.out.println("[Debug] - " + this.getClass().getName() + " - sendValueToRCX(): Sending value complete");
    				}

    				if(isDebug) {
    					System.out.println("[Debug] - " + this.getClass().getName() + " - sendValueToRCX(): Waiting for response from RCX, timeout: " + timeout + " ms");
    				}

    				// Wait for an ACK reply till the timeout is up (this timeout also throws an IOException when happening)
    				this.port.setTimeOut(this.timeout);
    				receivedInt = this.dInputStream.readInt();
    				this.port.setTimeOut(0);

    				if(isDebug) {
    					System.out.println("[Debug] - " + this.getClass().getName() + " - sendValueToRCX(): Received response from RCX: " + receivedInt);
    				}

    			} catch (IOException ex) {

    				// Notify the user of the timeout
    				System.err.println("[Error] - " + this.getClass().getName() + " - sendValueToRCX(): Reading from input stream timed out !");

    				// A timeout occured, raise the retry counter and resend
    				this.retryCounter++;
    			}
    		}

    		// Check if we stopped due to reaching the maximum number of retries
    		if(this.retryCounter > this.numberOfRetries) {
    			// Error, reset everything
    			System.err.println("[Error] - " + this.getClass().getName() + " - sendValueToRCX(): Too many retries! - Lost Connection ! Last Message Lost !");

    			// Throw the exception so that the super class get's notified of that error
    			throw new LostConnectionException();
    		}

    		if(isDebug) {
    			System.out.println("[Debug] - " + this.getClass().getName() + " - sendValueToRCX(): Everything successfull !");
    		}
    		return true;
    	}

    	/**
    	 * Used to send an ACK to the RCX
    	 * @throws LostConnectionException, gets thrown in case that the number of timeouts reaches the maximum number of retries
    	 */
    	private boolean sendACKToRCX() throws LostConnectionException {

    		if(isDebug) {
    			System.out.println("[Debug] - " + this.getClass().getName() + " - sendACKToRCX(): Sending ACK to RCX");
    		}

    		// Send the opcode till the RCX responds with an ACK
    		this.receivedIntACK = -1;
    		this.retryCounterACK = 0;

    		while(receivedIntACK != IRProtocol.OPCODE_MESSAGE_ACK && retryCounterACK <= this.numberOfRetries) {

    			try {
    				if(isDebug) {
    					System.out.println("[Debug] - " + this.getClass().getName() + " - sendACKToRCX(): Starting to send ACK");
    				}

    				// Send the value to the RCX with a timeout
    				this.writeToOutputStream(IRProtocol.OPCODE_MESSAGE_ACK);

    				// We won't get a reply to an ACK opcode we sent, so make sure we leave the while(...)
    				this.receivedIntACK = IRProtocol.OPCODE_MESSAGE_ACK;

    				if(isDebug) {
    					System.out.println("[Debug] - " + this.getClass().getName() + " - sendACKToRCX(): Sending ACK complete");
    				}

    			} catch (IOException ex) {

    				// Notify the user of the timeout
    				System.err.println("[Error] - " + this.getClass().getName() + " - sendACKToRCX(): Reading from input stream timed out !");

    				// A timeout occured, raise the retry counter and resend
    				this.retryCounterACK++;
    			}
    		}

    		// Check if we stopped due to reaching the maximum number of retries
    		if(this.retryCounterACK > this.numberOfRetries) {
    			// Error, reset everything
    			System.err.println("[Error] - " + this.getClass().getName() + " - sendACKToRCX(): Too many retries! - Lost Connection ! Last Message Lost !");

    			// Throw the exception so that the super class get's notified of that error
    			throw new LostConnectionException();
    		}

    		if(isDebug) {
    			System.out.println("[Debug] - " + this.getClass().getName() + " - sendACKToRCX(): Everything successfull !");
    		}
    		return true;
    	}

    	/**
    	 * Subclass used to create a small subthread to write a message onto the output stream
    	 * so that this subthread can be killed after a certain timeout.
    	 *
    	 * The "outputStream.write(int)" message does not support a timeout.
    	 */
    	class SendMessageHandler implements Runnable {

    		/* The value which should be written onto the output stream */
    		private int value = -1;

    		/**
    		 * Construct a new SendMessageHandler
    		 * @param value, the value to send to the RCX
    		 */
    		public SendMessageHandler(int value) {
    			this.value = value;

    			if(isDebug) {
    				System.out.println("[Debug] - " + this.getClass().getName() + " - constructor(): Creation complete");
    			}
    		}

    		/**
    		 * Run method, called when the Thread starts
    		 */
    		public void run() {

    			if(isDebug) {
    				System.out.println("[Debug] - " + this.getClass().getName() + " - run(): Thread started, trying to write value onto stream: " + this.value);
    			}

    			try {
    				dOutputStream.writeInt(this.value);
    				dOutputStream.flush();

    				if(isDebug) {
    					System.out.println("[Debug] - " + this.getClass().getName() + " - run(): Writing value to stream successfull");
    				}

    				// Set the boolean so that the main class knows that no timeout took place
    				isTimeout = false;

    			} catch (IOException e) {
    				System.err.println("[Error] - " + this.getClass().getName() + " - run(): Error while writing to outputStream: " + e);
    			}
    		}
    	}

    	/**
    	 * Subclass taking care of receiving messages
    	 */
    	class ReceiveMessageHandler {

    		/* Variables to store data */
    		private int readValue = -1;

    		/* Tells us if we already have the full message or not */
    		private boolean messageIncomplete = true;

    		/* The message object we reply with when reading */
    		private IRMessage message = new IRMessage();

    		/**
    		 * Construct a new ReceiveMessageHandler
    		 */
    		public ReceiveMessageHandler() {
    			if(isDebug) {
    				System.out.println("[Debug] - " + this.getClass().getName() + " - constructor(): Creation complete");
    			}
    		}

    		/**
    		 * Start to wait for a message from the RCX.
    		 * This method blocks the execution till it received a complete full message
    		 * from the RCX.
    		 * @return IRMessage, the fully received message from the RCX
    		 * @throws LostConnectionException, thrown in case the number of timeouts reached the maximum number of retries
    		 */
    		public IRMessage read() throws LostConnectionException {

    			try {
    				// Set the boolean to indicate that the message is not complete yet
    				this.messageIncomplete = true;

    				// Reset the values to get ready for a new message
    				this.reset();

    				// Receive packets till the message is complete
    				while(this.messageIncomplete) {

    					// Read a value from the input stream (no timeout)
    					port.setTimeOut(0);
    					readValue = dInputStream.readInt();

    					if(isDebug) {
    						System.out.println("[Debug] - " + this.getClass().getName() + " - read(): Received a value from the RCX: " + readValue);
    					}

    					// Acknowledge this message in case it was no ACK we received
    					if(readValue != IRProtocol.OPCODE_MESSAGE_ACK) {

    						// No ACK, send one
    						boolean sendACKSuccess = sendACKToRCX();

    						if(!sendACKSuccess) {
    							System.err.println("Sending ACK was not successfull !");
    						}
    					}

    					// Divide opcode and value
    					int distance = readValue%1000;
    					int opcode = readValue - distance;

    					// Build the IRMessage
    					this.message = new IRMessage(opcode, distance);

    					this.messageIncomplete = false;
    				}

    				if(isDebug) {
    					System.out.println("[Debug] - " + this.getClass().getName() + " - read(): Received Message is now complete (opcode / value): " + this.message.getOpcode() + " / " + this.message.getValue());
    				}

    				// Message is complete, return it to the caller of the method
    				return this.message;

    			} catch (IOException e) {
    				// Error while receiving a message
    				System.out.println("[Error] - " + this.getClass().getName() + " - read(): Error while received a message: " + e);

    				return null;
    			}
    		}

    		/**
    		 * Reset everything
    		 * Used when we know that we will receive the whole last message again
    		 */
    		private void reset() {

    			if(isDebug) {
    				System.out.println("[Debug] - " + this.getClass().getName() + " - reset(): Resetting everything !");
    			}

    			this.message.setOpcode(-1);
    			this.message.setValue(-1);
    		}
    	}

    	/**
    	 * Is this class currently in debug mode?
    	 * @return true if debug mode is active, false otherwise
    	 */
    	public boolean isDebug() {
    		return isDebug;
    	}

    	/**
    	 * Set the debug mode of this program
    	 * @param isDebug, true if debug messages should be shown, false otherwise
    	 */
    	public void setDebug(boolean isDebug) {
    		this.isDebug = isDebug;

    		if(isDebug) {
    			System.out.println("[Debug] - " + this.getClass().getName() + " - setDebug(): Debug mode enabled !");
    		}
    	}

    	/**
    	 * Returns how often we do a retry in case a timeout occurs during reading or writing
    	 * to the streams
    	 * @return Returns the numberOfRetries.
    	 */
    	public int getNumberOfRetries() {
    		return numberOfRetries;
    	}

    	/**
    	 * Set the number of retries how often the program will try to send a message after a timeout
    	 * @param numberOfRetries
    	 */
    	public void setNumberOfRetries(int numberOfRetries) {
    		this.numberOfRetries = numberOfRetries;

    		if(isDebug) {
    			System.out.println("[Debug] - " + this.getClass().getName() + " - setNumberOfRetries(): New number of timeout retries: " + this.numberOfRetries);
    		}
    	}

    	/**
    	 * Retrieve the time till a timeout during reading or writing occurs
    	 * @return Returns the timeout in milliseconds
    	 */
    	public int getTimeout() {
    		return timeout;
    	}

    	/**
    	 * Set the time till a timeout during reading or writing occurs
    	 * @param timeout, the new timeout in milliseconds
    	 */
    	public void setTimeout(int timeout) {
    		this.timeout = timeout;

    		if(isDebug) {
    			System.out.println("[Debug] - " + this.getClass().getName() + " - setTimeout(): New time till a timeout occurs: " + this.timeout);
    		}
    	}
    }

	/**
	 * Munich University of Applied Sciences, FB07 Computer Science
	 * FWP - Remotely Controlled Robot - 7.Semester
	 *
	 * Class Description:
	 * 		This class acts as IRMessage object which is assembeled when
	 * 		we want to send a message to the RCX/Proxy or when one has been received
	 *
	 * @author Thomas Woellert (thomas.woellert@t-online.de)
	 * @author Stefan Weissbach (weissbst@gmx.de)
	 * @author Willi Tremmel (WTremmel@gmx.de)
	 *
	 * System: Linux 2.6, Sun Java JDK 5.0
	 * Semestergroup: 7T
	 */

	public class IRMessage {

		/* The opcode */
		protected int opcode = -1;

		/* The additional value */
		protected int value = -1;

		/**
		 * Construct a new empty IRMessage
		 */
		public IRMessage() {
			this(-1, -1);
		}

		/**
		 * Constructa new IRMessage with the given opcode and value
		 * @param opcode, the opcode to store in the message
		 * @param value, the value to store in the message
		 */
		public IRMessage(int opcode, int value) {
			this.opcode = opcode;
			this.value = value;
		}

		/**
		 * Retrieve the stored opcode
		 * @return Returns the opcode.
		 */
		public int getOpcode() {
			return opcode;
		}

		/**
		 * Retrieve the stored value
		 * @return Returns the value.
		 */
		public int getValue() {
			return value;
		}

		/**
		 * Set a new opcode in the message
		 * @param opcode The opcode to set.
		 */
		public void setOpcode(int opcode) {
			this.opcode = opcode;
		}

		/**
		 * Set a new value in the message
		 * @param value The value to set.
		 */
		public void setValue(int value) {
			this.value = value;
		}
	}

}